from ..foundation import response_handler
from ..request import Request
from ..response import Response


class FrameworkProvider:
    def __init__(self, application):
        self.application = application

    def register(self):
        pass

    def boot(self):
        request = Request(self.application.make("environ"))
        request.app = self.application
        self.application.bind("request", request)
        self.application.bind("response", Response(self.application))



from .. import Middleware
from ...utils.str import random_string
from ...facades import Request, Session, Response


class SessionMiddleware(Middleware):
    def before(self, request, response):
        if not request.cookie("SESSID"):
            session_code = random_string(10)
            response.cookie("SESSID", session_code)
            request.cookie("SESSID", session_code)
        Session.start()
        request.app.make("response").with_input = self.with_input
        request.app.make("response").with_errors = self.with_errors
        request.app.make("request").session = Session
        return request

    def after(self, request, _):
        Session.save()
        return request

    def with_input(self):
        for key, value in Request.all().items():
            Session.flash(key, value)
        return Response

    def with_errors(self, errors):
        Session.flash("errors", errors)
        return Response



class HasColoredOutput:
    """Add level-colored output print functions to a class."""

    def success(self, message):
        print("\033[92m {0} \033[0m".format(message))

    def warning(self, message):
        print("\033[93m {0} \033[0m".format(message))

    def danger(self, message):
        print("\033[91m {0} \033[0m".format(message))

    def info(self, message):
        return self.success(message)



from ..cookies import CookieJar
from ..headers import HeaderBag, Header
from ..input import InputBag
import re
import tldextract
from .validation import ValidatesRequest


class Request(ValidatesRequest):
    def __init__(self, environ):
        """Request class constructor.

        Initializes several properties and sets various methods
        depending on the environtment.

        Keyword Arguments:
            environ {dictionary} -- WSGI environ dictionary. (default: {None})
        """
        self.environ = environ
        self.cookie_jar = CookieJar()
        self.header_bag = HeaderBag()
        self.input_bag = InputBag()
        self.params = {}
        self._user = None
        self.load()

    def load(self):
        self.cookie_jar.load(self.environ.get("HTTP_COOKIE", ""))
        self.header_bag.load(self.environ)
        self.input_bag.load(self.environ)

    def load_params(self, params=None):
        if not params:
            params = {}

        self.params.update(params)

    def param(self, param, default=""):
        return self.params.get(param, default)

    def get_path(self):
        return self.environ.get("PATH_INFO")

    def get_back_path(self):
        return self.input("__back") or self.get_path()

    def get_request_method(self):
        return self.environ.get("REQUEST_METHOD")

    def input(self, name, default=False):
        """Get a specific input value.

        Arguments:
            name {string} -- Key of the input data

        Keyword Arguments:
            default {string} -- Default value if input does not exist (default: {False})
            clean {bool} -- Whether or not the return value should be
                            cleaned (default: {True})

        Returns:
            string
        """
        name = str(name)

        return self.input_bag.get(name, default=default)

    def cookie(self, name, value=None):
        if value is None:
            cookie = self.cookie_jar.get(name)
            if not cookie:
                return
            return cookie.value

        return self.cookie_jar.add(name, value)

    def delete_cookie(self, name):
        self.cookie_jar.delete(name)
        return self

    def header(self, name, value=None):
        if value is None:
            header = self.header_bag.get(name)
            if not header:
                return
            return header.value
        else:
            return self.header_bag.add(Header(name, value))

    def all(self):
        return self.input_bag.all_as_values()

    def only(self, *inputs):
        return self.input_bag.only(*inputs)

    def is_not_safe(self):
        """Check if the current request is not a get request.

        Returns:
            bool
        """
        if not self.get_request_method() in ("GET", "OPTIONS", "HEAD"):
            return True

        return False

    def user(self):
        return self._user

    def set_user(self, user):
        self._user = user
        return self

    def remove_user(self):
        self._user = None
        return self

    def contains(self, route):
        if not route.startswith("/"):
            route = "/" + route

        regex = re.compile(route.replace("*", "[a-zA-Z0-9_]+"))

        return regex.match(self.get_path())

    def get_subdomain(self, exclude_www=True):
        url = tldextract.extract(self.get_host())
        if url.subdomain == "" or (
            url.subdomain and exclude_www and url.subdomain == "www"
        ):
            return None

        return url.subdomain

    def get_host(self):
        return self.environ.get("HTTP_HOST")



from inspect import isclass

from ..response import Response
from ..facades import Response as ResponseFacade
from .Provider import Provider
from ..routes import Route
from ..pipeline import Pipeline


class RouteProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        # Register the routes?
        Route.set_controller_locations(self.application.make("controllers.location"))

    def boot(self):
        router = self.application.make("router")
        request = self.application.make("request")
        response = self.application.make("response")

        route = router.find(
            request.get_path(), request.get_request_method(), request.get_subdomain()
        )

        # Run before middleware

        Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="before",
        )
        if route:
            request.load_params(route.extract_parameters(request.get_path()))
            self.application.make("middleware").run_route_middleware(
                route.list_middleware, request, response, callback="before"
            )
            exception = None

            try:
                data = route.get_response(self.application)
                if isinstance(data, Response) or (
                    isclass(data) and issubclass(data, ResponseFacade)
                ):
                    pass
                else:
                    response.view(data)
            except Exception as e:
                exception = e
                self.application.make("middleware").run_route_middleware(
                    route.list_middleware, request, response, callback="after"
                )
                if exception:
                    raise exception

            self.application.make("middleware").run_route_middleware(
                route.list_middleware, request, response, callback="after"
            )

        else:
            response.view("route not found", status=404)

        Pipeline(request, response).through(
            self.application.make("middleware").get_http_middleware(),
            handler="after",
        )



import re
import os

from ..utils.str import modularize, removeprefix
from ..exceptions import InvalidRouteCompileException
from ..facades import Loader
from ..controllers import Controller
from ..exceptions import LoaderNotFound


class HTTPRoute:
    def __init__(
        self,
        url,
        controller=None,
        request_method=["get"],
        name=None,
        compilers=None,
        controllers_locations=["app.http.controllers"],
        controller_bindings=[],
        **options,
    ):
        if not url.startswith("/"):
            url = "/" + url

        self.url = url
        self.controllers_locations = controllers_locations
        self.controller = controller
        self.controller_class = None
        self.controller_method = None
        self._domain = None
        self._name = name
        self.request_method = [x.lower() for x in request_method]
        self.list_middleware = []
        self.e = None
        self.compilers = compilers or {}
        self._find_controller(controller)
        self.controller_bindings = controller_bindings
        self.compile_route_to_regex()

    def match(self, path, request_method, subdomain=None):

        route_math = (
            re.match(self._compiled_regex, path)
            or re.match(self._compiled_regex_end, path)
        ) and request_method.lower() in self.request_method

        domain_match = subdomain == self._domain

        return route_math and domain_match

    def get_name(self):
        return self._name

    def matches(self, path):
        return re.match(self._compiled_regex, path) or re.match(
            self._compiled_regex_end, path
        )

    def match_name(self, name):
        return name == self._name

    def name(self, name):
        self._name = name
        return self

    def domain(self, subdomain):
        self._domain = subdomain
        return self

    def to_url(self, parameters={}):

        # Split the url into a list
        split_url = self.url.split("/")

        # Start beginning of the new compiled url
        compiled_url = "/"

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if "@" in url:
                    url = url.replace("@", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters[url]) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                elif "?" in url:
                    url = url.replace("?", "").split(":")[0]
                    if isinstance(parameters, dict):
                        compiled_url += str(parameters.get(url, "/")) + "/"
                    elif isinstance(parameters, list):
                        compiled_url += str(parameters.pop(0)) + "/"
                else:
                    compiled_url += url + "/"

        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith("/"):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if "//" in compiled_url:
            compiled_url = compiled_url.replace("//", "/")

        return compiled_url

    def _find_controller(self, controller):
        """Find the controller to attach to the route. Look for controller (str or class) in all
        specified controllers_location.

        Arguments:
            controller {string|object} -- String or object controller to search for.

        Returns:
            None
        """
        if controller is None:
            return None
        # If the output specified is a string controller e.g. "WelcomeController@show"
        elif isinstance(controller, str):
            if "@" in controller:
                controller_path, controller_method_str = controller.split("@")
            else:
                controller_path = controller
                controller_method_str = "__call__"

            controller_path = modularize(controller_path).split(".")
            if len(controller_path) > 1:
                controller_name = controller_path.pop()
                prefix_path = ".".join(controller_path)
            else:
                controller_name = controller_path[0]
                prefix_path = ""
            # build a list of all locations where the controller can be found
            # if the controller is defined such as auth.WelcomeController, append the prefix path to
            # the locations
            locations = list(
                map(
                    lambda loc: f"{loc}.{removeprefix(prefix_path, loc)}"
                    if prefix_path
                    else loc,
                    self.controllers_locations,
                )
            )
            try:
                self.controller_class = Loader.find(
                    Controller, locations, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")
        # Else it's a controller instance, we don't have to find it, just get the class
        else:
            if "." in controller.__qualname__:
                controller_name, controller_method_str = controller.__qualname__.split(
                    "."
                )
            else:
                controller_name = controller.__qualname__
                controller_method_str = "__call__"
            try:
                self.controller_class = Loader.get_object(
                    controller.__module__, controller_name, raise_exception=True
                )
            except LoaderNotFound as e:
                self.e = e
                print("\033[93mTrouble importing controller!", str(e), "\033[0m")

        # Set the controller method on class. This is a string
        self.controller_method = controller_method_str

    def get_response(self, app=None):
        # Resolve Controller Constructor
        if self.e:
            print(
                "\033[93mCannot find controller {}. Did you create this one?".format(
                    self.controller
                ),
                "\033[0m",
            )
            raise SyntaxError(str(self.e))

        if app:
            controller = app.resolve(self.controller_class, *self.controller_bindings)
            # resolve route parameters
            params = self.extract_parameters(app.make("request").get_path()).values()
            # Resolve Controller Method
            response = app.resolve(getattr(controller, self.controller_method), *params)
            return response

        return getattr(self.controller_class(), self.controller_method)()

    def middleware(self, *args):
        """Load a list of middleware to run.

        Returns:
            self
        """
        for arg in args:
            if arg and arg not in self.list_middleware:
                self.list_middleware.append(arg)

        return self

    def compile_route_to_regex(self):
        """Compile the given route to a regex string.

        Arguments:
            route {string} -- URI of the route to compile.

        Returns:
            string -- Compiled URI string.
        """
        # Split the route
        split_given_route = self.url.split("/")
        # compile the provided url into regex
        url_list = []
        regex = "^"
        for regex_route in split_given_route:
            if "@" in regex_route:
                if ":" in regex_route:
                    try:
                        regex += self.compilers[regex_route.split(":")[1]]
                    except KeyError:
                        raise InvalidRouteCompileException(
                            'Route compiler "{}" is not an available route compiler. '
                            "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                regex_route.split(":")[1]
                            )
                        )

                else:
                    regex += self.compilers["default"]

                regex += r"\/"

                # append the variable name passed @(variable):int to a list
                url_list.append(regex_route.replace("@", "").split(":")[0])
            elif "?" in regex_route:
                # Make the preceding token match 0 or more
                regex += "?"

                if ":" in regex_route:

                    try:
                        regex += self.compilers[regex_route.split(":")[1]] + "*"
                    except KeyError:
                        if self.request:
                            raise InvalidRouteCompileException(
                                'Route compiler "{}" is not an available route compiler. '
                                "Verify you spelled it correctly or that you have added it using the compile() method.".format(
                                    regex_route.split(":")[1]
                                )
                            )
                        self._compiled_regex = None
                        self._compiled_regex_end = None
                        return

                else:
                    regex += self.compilers["default"] + "*"

                regex += r"\/"

                url_list.append(regex_route.replace("?", "").split(":")[0])
            else:
                regex += regex_route + r"\/"

        self.url_list = url_list
        regex += "$"
        self._compiled_regex = re.compile(regex.replace(r"\/$", r"$"))
        self._compiled_regex_end = re.compile(regex)

        return regex

    def extract_parameters(self, path):
        if not self.url_list:
            return {}

        if (not path.endswith("/")) or path == "/":
            matching_regex = self._compiled_regex
        else:
            matching_regex = self._compiled_regex_end
        return dict(zip(self.url_list, matching_regex.match(path).groups()))



from .CommandCapsule import CommandCapsule
from .AuthCommand import AuthCommand
from .TinkerCommand import TinkerCommand
from .KeyCommand import KeyCommand
from .ServeCommand import ServeCommand
from .QueueWorkCommand import QueueWorkCommand
from .QueueRetryCommand import QueueRetryCommand
from .QueueTableCommand import QueueTableCommand
from .QueueFailedCommand import QueueFailedCommand
from .MakeControllerCommand import MakeControllerCommand
from .MakeJobCommand import MakeJobCommand
from .MakeMailableCommand import MakeMailableCommand
from .MakeProviderCommand import MakeProviderCommand



from .middleware_capsule import MiddlewareCapsule
from .middleware import Middleware
from .route.VerifyCsrfToken import VerifyCsrfToken
from .route.SessionMiddleware import SessionMiddleware
from .route.EncryptCookies import EncryptCookies
from .route.LoadUserMiddleware import LoadUserMiddleware



from tests import TestCase
from src.masonite.foundation import Application
import os
from src.masonite.response import Response
from src.masonite.routes import Router, Route


class TestResponseRedirect(TestCase):
    def setUp(self):
        application = Application(os.getcwd())
        application.bind("router", Router(Route.get("/", None).name("home")))
        self.response = Response(application)

    def test_redirect(self):
        self.response.redirect("/")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/")

    def test_redirect_to_route_named_route(self):
        self.response.redirect(name="home")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/")

    def test_redirect_to_url(self):
        self.response.redirect(url="/login")
        self.assertEqual(self.response.get_status(), 302)
        self.assertEqual(self.response.header_bag.get("Location").value, "/login")



from src.masonite.tests import TestCase
from src.masonite.routes import Route


class TestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.addRoutes(
            Route.set_controller_locations("tests.integrations.controllers").get(
                "/", "WelcomeController@show"
            ),
            Route.set_controller_locations("tests.integrations.controllers").post(
                "/", "WelcomeController@show"
            ),
        )



from tests import TestCase

from src.masonite.utils.http import generate_wsgi
from src.masonite.request import Request


class TestRequest(TestCase):
    def setUp(self):
        self.request = Request(generate_wsgi(path="/test"))

    def test_request_no_input_returns_false(self):
        self.assertEqual(self.request.input("notavailable"), False)

    def test_request_can_get_string_value(self):
        self.request.input_bag.query_string = {"test": "value"}
        self.assertEqual(self.request.input("test"), "value")

    def test_request_can_get_list_value(self):
        self.request.input_bag.query_string = {"test": ["foo", "bar"]}
        self.assertEqual(self.request.input("test"), ["foo", "bar"])

    def test_request_can_get_float_value(self):
        self.request.input_bag.query_string = {"test": 3.1415926}
        self.assertEqual(self.request.input("test"), 3.1415926)



class Session:
    def __init__(self, application, driver_config=None):
        self.application = application
        self.drivers = {}
        self._driver = None
        self.driver_config = driver_config or {}
        self.options = {}
        self.data = {}
        self.added = {}
        self.flashed = {}
        self.deleted = []
        self.deleted_flashed = []

    def add_driver(self, name, driver):
        self.drivers.update({name: driver})

    def driver(self, driver):
        return self.drivers[driver]

    def set_configuration(self, config):
        self.driver_config = config
        return self

    def get_driver(self, name=None):
        if name is None:
            return self.drivers[self.driver_config.get("default")]
        return self.drivers[name]

    def get_config_options(self, driver=None):
        if driver is None:
            return self.driver_config[self.driver_config.get("default")]

        return self.driver_config.get(driver, {})

    # Start of methods
    def start(self, driver=None):
        self.data = {}
        self.added = {}
        self.flashed = {}
        self.deleted = []
        self.deleted_flashed = []
        started_data = self.get_driver(name=driver).start()
        self.data = started_data.get("data")
        self.flashed = started_data.get("flashed")
        return self

    def get_data(self):
        data = self.data
        data.update(self.added)
        data.update(self.flashed)
        for deleted in self.deleted:
            if deleted in data:
                data.pop(deleted)
        for deleted in self.deleted_flashed:
            if deleted in data:
                data.pop(deleted)
        return data

    def save(self, driver=None):
        return self.get_driver(name=driver).save(
            added=self.added,
            deleted=self.deleted,
            flashed=self.flashed,
            deleted_flashed=self.deleted_flashed,
        )

    def set(self, key, value):
        return self.added.update({key: value})

    def increment(self, key, count=1):
        return self.set(key, str(int(self.get(key)) + count))

    def decrement(self, key, count=1):
        return self.set(key, str(int(self.get(key)) - count))

    def has(self, key):
        return key in self.get_data()

    def get(self, key):
        if key in self.flashed:
            value = self.flashed.get(key)
            self.flashed.pop(key)
            self.deleted_flashed.append(key)
            return value

        return self.get_data().get(key)

    def pull(self, key):
        key_value = self.get(key)
        self.delete(key)
        return key_value

    def flush(self):
        self.deleted += list(self.get_data().keys())

    def delete(self, key):
        self.deleted.append(key)
        if key in self.flashed:
            self.flashed.pop(key)

    def flash(self, key, value):
        """Add temporary data to the session.

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        self.flashed.update({key: value})

    def all(self):
        """Get all session data.

        Returns:
            dict
        """
        return self.get_data()



""" A ScheduleRunCommand Command """
import pendulum
import inspect
from cleo import Command

from ..Task import Task


class ScheduleRunCommand(Command):
    """
    Run the scheduled tasks
    schedule:run
        {--t|task=None : Name of task you want to run}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        return self.app.make("scheduler").run()



class CommandCapsule:
    def __init__(self, command_application):
        self.command_application = command_application
        self.commands = []

    def add(self, *commands):
        self.commands.append(commands)
        self.command_application.add_commands(*commands)
        return self

    def run(self):
        return self.command_application.run()



from tests import TestCase
from src.masonite.foundation import Application
import os
from src.masonite.response import Response


class TestResponseRedirect(TestCase):
    def setUp(self):
        application = Application(os.getcwd())
        self.response = Response(application)

    def test_download(self):
        self.response.download(
            "invoice", "tests/integrations/storage/invoice.pdf", force=True
        )
        self.assertTrue(self.response.header("Content-Disposition"))
        self.assertEqual(
            self.response.header("Content-Disposition"),
            'attachment; filename="invoice.pdf"',
        )



import pendulum

from tests import TestCase
from tests.integrations.controllers.WelcomeController import WelcomeController
from masoniteorm.models import Model
from src.masonite.routes import Route
from src.masonite.authentication import Authenticates


class User(Model, Authenticates):
    pass


class CustomTestResponse:
    def assertCustom(self):
        assert 1
        return self


class OtherCustomTestResponse:
    def assertOtherCustom(self):
        assert 2
        return self


class TestTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            Route.get("/", "WelcomeController@show").name("home"),
        )

    def tearDown(self):
        super().tearDown()
        self.restoreTime()

    def test_add_routes(self):
        self.assertEqual(len(self.application.make("router").routes), 1)
        self.addRoutes(
            Route.get("/test", "WelcomeController@show").name("test"),
        )
        self.assertEqual(len(self.application.make("router").routes), 2)

    def test_use_custom_test_response(self):
        self.application.make("tests.response").add(
            CustomTestResponse, OtherCustomTestResponse
        )
        # can use default assertions and custom from different classes
        self.get("/").assertContains("Welcome").assertCustom().assertOtherCustom()

    def test_fake_time(self):
        given_date = pendulum.datetime(2015, 2, 5)
        self.fakeTime(given_date)
        self.assertEqual(pendulum.now(), given_date)
        self.restoreTime()
        self.assertNotEqual(pendulum.now(), given_date)

    def test_fake_time_tomorrow(self):
        tomorrow = pendulum.tomorrow()
        self.fakeTimeTomorrow()
        self.assertEqual(pendulum.now(), tomorrow)

    def test_fake_time_yesterday(self):
        yesterday = pendulum.yesterday()
        self.fakeTimeYesterday()
        self.assertEqual(pendulum.now(), yesterday)

    def test_fake_time_in_future(self):
        real_now = pendulum.now()
        self.fakeTimeInFuture(10)
        self.assertEqual(pendulum.now().diff(real_now).in_days(), 10)
        self.assertGreater(pendulum.now(), real_now)

        self.fakeTimeInFuture(1, "months")
        self.assertEqual(pendulum.now().diff(real_now).in_months(), 1)

    # def test_fake_time_in_past(self):
    #     real_now = pendulum.now()
    #     self.fakeTimeInPast(10)
    #     self.assertEqual(pendulum.now().diff(real_now).in_days(), 10)
    #     self.assertLess(pendulum.now(), real_now)

    #     self.fakeTimeInPast(3, "hours")
    #     self.assertEqual(real_now.hour - pendulum.now().hour, 3)


class TestTestingAssertions(TestCase):
    def setUp(self):
        super().setUp()
        self.setRoutes(
            Route.get("/", "WelcomeController@show").name("home"),
            Route.get("/test", "WelcomeController@show").name("test"),
            Route.get("/view", "WelcomeController@view").name("view"),
            Route.get("/view-context", "WelcomeController@view_with_context").name(
                "view_with_context"
            ),
            Route.get("/test-404", "WelcomeController@not_found").name("not_found"),
            Route.get("/test-creation", "WelcomeController@create").name("create"),
            Route.get("/test-unauthorized", "WelcomeController@unauthorized").name(
                "unauthorized"
            ),
            Route.get("/test-forbidden", "WelcomeController@forbidden").name(
                "forbidden"
            ),
            Route.get("/test-empty", "WelcomeController@empty").name("empty"),
            Route.get(
                "/test-response-header", "WelcomeController@response_with_headers"
            ),
            Route.get("/test-redirect-1", "WelcomeController@redirect_url"),
            Route.get("/test-redirect-2", "WelcomeController@redirect_route"),
            Route.get("/test-redirect-3", "WelcomeController@redirect_route_params"),
            Route.get("/test/@id", "WelcomeController@with_params").name("test_params"),
            Route.get("/test-json", "WelcomeController@json").name("json"),
            Route.get("/test-session", "WelcomeController@session").name("session"),
            Route.get(
                "/test-session-errors", "WelcomeController@session_with_errors"
            ).name("session"),
            Route.get("/test-session-2", "WelcomeController@session2").name("session2"),
            Route.get("/test-authenticates", "WelcomeController@auth").name("auth"),
        )

    def test_assert_contains(self):
        self.get("/").assertContains("Welcome")
        self.get("/").assertNotContains("hello")

    def test_assert_is_named(self):
        self.get("/test").assertIsNamed("test")
        self.get("/test").assertIsNotNamed("welcome")

    def test_assert_not_found(self):
        self.get("/test-404").assertNotFound()

    def test_assert_is_status(self):
        self.get("/test").assertIsStatus(200)

    def test_assert_ok(self):
        self.get("/test").assertOk()

    def test_assert_created(self):
        self.get("/test-creation").assertCreated()

    def test_assert_unauthorized(self):
        self.get("/test-unauthorized").assertUnauthorized()

    def test_assert_forbidden(self):
        self.get("/test-forbidden").assertForbidden()

    def test_assert_no_content(self):
        self.get("/test-empty").assertNoContent()

    def test_assert_cookie(self):
        self.withCookies({"test": "value"}).get("/").assertCookie("test")

    def test_assert_cookie_value(self):
        self.withCookies({"test": "value"}).get("/").assertCookie("test", "value")

    def test_assert_cookie_missing(self):
        self.get("/").assertCookieMissing("test")

    def test_assert_plain_cookie(self):
        self.withCookies({"test": "value"}).get("/").assertPlainCookie("test")

    def test_assert_has_header(self):
        self.get("/test-response-header").assertHasHeader("TEST")
        self.get("/test-response-header").assertHasHeader("TEST", "value")

    def test_assert_header_missing(self):
        self.get("/").assertHeaderMissing("X-Test")

    def test_assert_request_with_headers(self):
        request = self.withHeaders({"X-TEST": "value"}).get("/").request
        assert request.header("X-Test") == "value"

    def test_assert_redirect_to_url(self):
        self.get("/test-redirect-1").assertRedirect("/")

    def test_assert_redirect_to_route(self):
        self.get("/test-redirect-2").assertRedirect(name="test")
        self.get("/test-redirect-3").assertRedirect(
            name="test_params", params={"id": 1}
        )

    def test_assert_session_has(self):
        self.get("/test-session").assertSessionHas("key")
        self.get("/test-session").assertSessionHas("key", "value")

    def test_assert_session_has_errors(self):
        self.get("/test-session-errors").assertSessionHasErrors()
        self.get("/test-session-errors").assertSessionHasErrors(["email"])
        self.get("/test-session-errors").assertSessionHasErrors(["email", "password"])

    def test_assert_session_has_no_errors(self):
        self.get("/test-session").assertSessionHasNoErrors()
        self.get("/test-session-errors").assertSessionHasNoErrors(["name"])

    def test_assert_session_missing(self):
        self.get("/").assertSessionMissing("some_test_key")

    def test_assert_view_is(self):
        self.get("/view").assertViewIs("welcome")

    def test_assert_view_has(self):
        self.get("/view-context").assertViewHas("count")
        self.get("/view-context").assertViewHas("count", 1)
        self.get("/view-context").assertViewHas("users", ["John", "Joe"])
        self.get("/view-context").assertViewHas("other_key.nested")
        self.get("/view-context").assertViewHas("other_key.nested", 1)

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHas("not_in_view")
        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHas("not_in_view", 3)

    def test_assert_view_helpers_raise_error_if_not_rendering_a_view(self):
        # json response
        with self.assertRaises(ValueError):
            self.get("/test-json").assertViewIs("test")
        # string response
        self.get("/test").assertViewIs("welcome")

    def test_assert_view_has_exact(self):
        self.get("/view-context").assertViewHasExact(["users", "count", "other_key"])
        self.get("/view-context").assertViewHasExact(
            {"count": 1, "users": ["John", "Joe"], "other_key": {"nested": 1}}
        )

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHasExact(
                ["users", "count", "not in data"]
            )

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewHasExact({"count": 1})

    def test_assert_view_missing(self):
        self.get("/view-context").assertViewMissing("not in data")

        with self.assertRaises(AssertionError):
            self.get("/view-context").assertViewMissing("users")

    def test_assert_guest(self):
        self.get("/test").assertGuest()

    # def test_assert_authenticated(self):
    #     self.get("/test-authenticates").assertAuthenticated()

    def test_assert_authenticated_as(self):
        self.make_request()
        self.application.make("auth").guard("web").attempt(
            "idmann509@gmail.com", "secret"
        )
        user = User.find(1)
        self.get("/test").assertAuthenticatedAs(user)

    def test_assert_has_controller(self):
        self.get("/test").assertHasController("WelcomeController@show")
        self.get("/test").assertHasController(WelcomeController)

    def test_assert_route_has_parameter(self):
        self.get("/test/3").assertRouteHasParameter("id")
        with self.assertRaises(AssertionError):
            self.get("/test/3").assertRouteHasParameter("key")
        # self.get("/test/3").assertRouteHasParameter("id", 3)
        # with self.assertRaises(AssertionError):
        #     self.get("/test/3").assertRouteHasParameter("id", 4)

    def test_assert_has_route_middleware(self):
        self.get("/test").assertHasRouteMiddleware("web")

    def test_assert_has_http_middleware(self):
        # TODO: add one for testing purposes
        # self.get("/test").assertHasHttpMiddleware()
        pass

    def test_assert_json(self):
        self.get("/test-json").assertJson({"key": "value"})
        # works also in a nested path
        self.get("/test-json").assertJson(
            {"other_key": {"nested": 1, "nested_again": {"a": 1, "b": 2}}}
        )

    def test_json_assertions_fail_when_response_not_json(self):
        with self.assertRaises(ValueError):
            self.get("/view").assertJson({"key": "value"})

    def test_assert_json_path(self):
        self.get("/test-json").assertJsonPath("key2", [1, 2])
        self.get("/test-json").assertJsonPath("other_key.nested", 1)
        self.get("/test-json").assertJsonPath("other_key.nested_again.b", 2)
        self.get("/test-json").assertJsonPath(
            "other_key.nested_again", {"a": 1, "b": 2}
        )

    def test_assert_json_count(self):
        self.get("/test-json").assertJsonCount(3)
        self.get("/test-json").assertJsonCount(2, key="other_key")

    def test_assert_json_exact(self):
        self.get("/test-json").assertJsonExact(
            {
                "key": "value",
                "key2": [1, 2],
                "other_key": {
                    "nested": 1,
                    "nested_again": {"a": 1, "b": 2},
                },
            }
        )

    def test_assert_json_missing(self):
        self.get("/test-json").assertJsonMissing("key3")
        self.get("/test-json").assertJsonMissing("some_key.nested")
        with self.assertRaises(AssertionError):
            self.get("/test-json").assertJsonMissing("other_key.nested")

    def test_assert_database_count(self):
        self.assertDatabaseCount("users", 1)

    def test_assert_database_has(self):
        self.assertDatabaseHas("users", {"name": "Joe"})

    def test_assert_database_missing(self):
        self.assertDatabaseMissing(
            "users", {"name": "John", "email": "john@example.com"}
        )



"""New Controller Command."""
from cleo import Command
import inflection
import os

from ..utils.location import controllers_path
from ..utils.filesystem import get_module_dir, render_stub_file


class MakeControllerCommand(Command):
    """
    Creates a new controller class.

    controller
        {name : Name of the controller}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        if not name.endswith("Controller"):
            name += "Controller"

        content = render_stub_file(self.get_controllers_path(), name)

        filename = f"{name}.py"
        with open(controllers_path(filename), "w") as f:
            f.write(content)

        self.info(f"Controller Created ({controllers_path(filename, absolute=False)})")

    def get_template_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/templates/")

    def get_controllers_path(self):
        return os.path.join(
            get_module_dir(__file__), "../stubs/controllers/Controller.py"
        )



import builtins

from .Provider import Provider
from ..exceptions import ExceptionHandler, DumpExceptionHandler, DD
from ..configuration import config


class ExceptionProvider(Provider):
    def __init__(self, application):
        self.application = application

    def register(self):
        handler = ExceptionHandler(self.application).set_options(config("exceptions"))
        builtins.dd = DD(self.application).dump
        self.application.bind("exception_handler", handler)
        self.application.bind(
            "DumpExceptionHandler", DumpExceptionHandler(self.application)
        )

    def boot(self):
        pass



"""New Mailable Command."""
from cleo import Command
import inflection
import os

from ..utils.filesystem import make_directory, render_stub_file, get_module_dir
from ..utils.str import as_filepath
from ..utils.location import base_path


class MakeMailableCommand(Command):
    """
    Creates a new mailable class.

    mailable
        {name : Name of the mailable}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = inflection.camelize(self.argument("name"))
        content = render_stub_file(self.get_mailables_path(), name)

        relative_filename = os.path.join(
            as_filepath(self.app.make("mailables.location")), name + ".py"
        )
        filepath = base_path(relative_filename)
        make_directory(filepath)

        with open(filepath, "w") as f:
            f.write(content)

        self.info(f"Mailable Created ({relative_filename})")

    def get_mailables_path(self):
        return os.path.join(get_module_dir(__file__), "../stubs/mailable/Mailable.py")



import os
from cleo import Application as CommandApplication

from .response_handler import response_handler
from ..commands import (
    TinkerCommand,
    CommandCapsule,
    KeyCommand,
    ServeCommand,
    QueueWorkCommand,
    QueueRetryCommand,
    QueueTableCommand,
    QueueFailedCommand,
    AuthCommand,
    MakeControllerCommand,
    MakeJobCommand,
    MakeMailableCommand,
    MakeProviderCommand,
)
from ..environment import LoadEnvironment
from ..middleware import MiddlewareCapsule
from ..routes import Router
from ..loader import Loader

from ..tests.HttpTestResponse import HttpTestResponse
from ..tests.TestResponseCapsule import TestResponseCapsule


class Kernel:
    def __init__(self, app):
        self.application = app

    def register(self):
        self.load_environment()
        self.register_framework()
        self.register_commands()
        self.register_testing()

    def load_environment(self):
        LoadEnvironment()

    def register_framework(self):
        self.application.set_response_handler(response_handler)
        self.application.use_storage_path(
            os.path.join(self.application.base_path, "storage")
        )
        self.application.bind("middleware", MiddlewareCapsule())
        self.application.bind(
            "router",
            Router(),
        )
        self.application.bind("loader", Loader())

    def register_commands(self):
        self.application.bind(
            "commands",
            CommandCapsule(CommandApplication("Masonite Version:", "4.0")).add(
                TinkerCommand(),
                KeyCommand(),
                ServeCommand(self.application),
                QueueWorkCommand(self.application),
                QueueRetryCommand(self.application),
                QueueFailedCommand(),
                QueueTableCommand(),
                AuthCommand(self.application),
                MakeControllerCommand(self.application),
                MakeJobCommand(self.application),
                MakeMailableCommand(self.application),
                MakeProviderCommand(self.application),
            ),
        )

    def register_testing(self):
        test_response = TestResponseCapsule(HttpTestResponse)
        self.application.bind("tests.response", test_response)



"""Starts Interactive Console Command."""
import code
import sys
from cleo import Command

from ..environment import env
from ..utils.collections import collect
from ..utils.structures import load, data_get
from ..utils.location import base_path, config_path
from ..helpers import optional, url
from ..facades import Loader


BANNER = """Masonite Python \033[92m {} \033[0m Console
This interactive console has the following things imported:
    -\033[92m app(container), \033[0m
    - Utils:\033[92m {}, \033[0m
    - Models:\033[92m {}, \033[0m

Type `exit()` to exit."""


class TinkerCommand(Command):
    """
    Run a python shell with the container pre-loaded.

    tinker
        {--i|ipython : Run a IPython shell}
    """

    def handle(self):
        from wsgi import application
        from masoniteorm.models import Model

        version = "{}.{}.{}".format(
            sys.version_info.major, sys.version_info.minor, sys.version_info.micro
        )
        models = Loader.find_all(Model, "tests/integrations/app")
        banner = BANNER.format(
            version,
            "env, optional, load, collect, url, asset, route, load, data_get, base_path, config_path",
            ",".join(models.keys()),
        )
        helpers = {
            "app": application,
            "env": env,
            "optional": optional,
            "collect": collect,
            "url": url.url,
            "asset": url.asset,
            "route": url.route,
            "load": load,
            "data_get": data_get,
            "base_path": base_path,
            "config_path": config_path,
        }
        context = {**helpers, **models}

        if self.option("ipython"):
            try:
                import IPython
            except ImportError:
                raise ModuleNotFoundError(
                    "Could not find the 'IPython' library. Run 'pip install ipython' to fix this."
                )
            from traitlets.config import Config

            c = Config()
            c.TerminalInteractiveShell.banner1 = banner
            IPython.start_ipython(argv=[], user_ns=context, config=c)
        else:
            code.interact(banner=banner, local=context)



import sys

import hupper
import waitress
from cleo import Command


class ServeCommand(Command):
    """
    Run the Masonite server.

    serve
        {--p|port=8000 : Specify which port to run the server}
        {--b|host=127.0.0.1 : Specify which ip address to run the server}
        {--r|reload : Make the server automatically reload on file changes}
        {--d|dont-reload : Make the server NOT automatically reload on file changes}
        {--i|reload-interval=1 : Number of seconds to wait to reload after changed are detected}
        {--l|live-reload : Make the server automatically refresh your web browser}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        if self.option("live-reload"):
            try:
                from livereload import Server
            except ImportError:
                raise ImportError(
                    "Could not find the livereload library. Install it by running 'pip install livereload==2.5.1'"
                )

            import glob

            server = Server(self.app)
            for filepath in glob.glob("resources/templates/**/*/"):
                server.watch(filepath)

            self.line("")
            self.info("Live reload server is starting...")
            self.info(
                "This will only work for templates. Changes to Python files may require a browser refresh."
            )
            self.line("")
            server.serve(
                port=self.option("port"),
                restart_delay=self.option("reload-interval"),
                liveport=5500,
                root=self.app.base_path,
                debug=True,
            )
            return

        reloader = hupper.start_reloader(self.app.make("server.runner"))

        # monitor an extra file
        reloader.watch_files([".env", self.app.get_storage_path()])


def main(args=sys.argv[1:]):
    from wsgi import application

    host = "127.0.0.1"
    port = "8000"
    if "--host" in args:
        host = args[args.index("--host") + 1]
    if "--port" in args:
        port = args[args.index("--host") + 1]
    if "-p" in args:
        port = args[args.index("-p") + 1]

    waitress.serve(application, host=host, port=port)
